/*  
 	HEX2BIN

	Utility to convert Intel Hex or Motorola S-Record files to a binary
	file. Only 16-bit hex files and a 64K address space are supported
  	(no segment records).

	Usage: HEX2BIN fname.ext [FillByte]
	Where: .ext = .hex or .s19
          FillByte = 00, FF, etc. (00 is default)
          Output file created is fname.bin      

	Written by Mike Douglas
	
	 Ver     Date             Desc
	----   ----------    ------------------------------------------------ 
	 1.0   01/04/2023		Initial version
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define FILENAME_LEN 81
#define INBUF_LEN 600
#define MEM_SIZE 64*1024

/* display_usage - Display how the program is used and then exit */

void display_usage (void)
{
	fprintf(stderr, "Usage: HEX2BIN fname.ext [FillByte]\n\n"
					    "Where: .ext = .hex or .s19\n"
					    "       FillByte = 00, FF, etc. (00 is default)\n"
					    "       Output file created is fname.bin\n");
	exit(1);
}

/* main */

int main (int argc, char *argv[])
{
	char	fileName[FILENAME_LEN];
	char	inBuf[INBUF_LEN];
	char	dataStr[INBUF_LEN];
	unsigned char *binImage;

	FILE	*inFile, *outFile;
	int	byteCount, dataAddr, recType, dataByte, checkSum, i;
	int	padByte = 0;
	int	lowAddr = MEM_SIZE-1;
	int	highAddr = 0;
	char	*bufPtr;
	int	sRecords = 0;
	
/* Process command line parameters */

	fprintf(stdout,"\nHEX2BIN v1.0\n\n");
	if ((argc < 2) || (argc > 3))
		display_usage();

	strcpy(fileName, argv[1]);
	if ((bufPtr = strrchr(fileName,'.')) == NULL) 
		display_usage();
	*bufPtr = '\0';

	if (tolower(bufPtr[1]) == 's')				/* .S19 for example */
		sRecords = 1;
	else if (tolower(bufPtr[1]) != 'h')			/* .HEX */
		display_usage();

	if (islower(bufPtr[1]))
		strcat(fileName, ".bin");					/* output file */
	else
		strcat(fileName, ".BIN");					/* output file */

	if (argc == 3)
		sscanf(argv[2], "%2x", &padByte);

	if ((binImage = malloc(MEM_SIZE)) == NULL) {
		fprintf(stderr,"Can't allocate image memory\n");
		exit(1);
	}
	memset(binImage, padByte, MEM_SIZE);

	if ((inFile = fopen(argv[1], "r")) == NULL) {
		fprintf(stderr,"Can't open file %s\n", argv[1]);
		exit(1);
	}

/* Read the hex or s-record file and put data into a memory image array */

	do {
		fgets(inBuf, INBUF_LEN, inFile);
		i = strlen(inBuf)-1;
		if (inBuf[i] == '\n')
			inBuf[i] = '\0';

		if (sRecords) {
			sscanf(inBuf, "S%1X%2x%4x%s", &recType, &byteCount, &dataAddr, dataStr);
			checkSum = byteCount + (dataAddr >> 8) + (dataAddr & 0xff) + 1;
			byteCount -= 3;		/* make byteCount = data bytes only */
			recType--;				/* make S1 match .hex record type 0 */
		}
		else {
			sscanf(inBuf, ":%2x%4x%2x%s", &byteCount, &dataAddr, &recType, dataStr);
			checkSum = byteCount + (dataAddr >> 8) + (dataAddr & 0xff) + recType;
		}
		bufPtr = dataStr;

		if ((recType == 0) && (byteCount > 0) && (dataAddr+byteCount <= MEM_SIZE)) {
			if (dataAddr+byteCount > highAddr)
				highAddr = dataAddr + byteCount;
			if (dataAddr < lowAddr)
				lowAddr = dataAddr;

			do {
				sscanf(bufPtr, "%2x", &dataByte);
				bufPtr += 2;
				binImage[dataAddr++] = dataByte;
				checkSum += dataByte;
			} 
			while (--byteCount != 0);

			sscanf(bufPtr, "%2x", &dataByte);			/* checksum byte */
			if (0 != ((checkSum+dataByte) & 0xff)) {
				fprintf(stderr,"Checksum error\n  %s\n", inBuf);
				exit(1);
			}
		} 
	} while (!feof(inFile));

/* Display results, write binary file */

	fprintf(stdout,"Low address  = %04X\n", lowAddr);
	fprintf(stdout,"High address = %04X\n", highAddr-1);
 	fprintf(stdout,"Fill Byte    = %02X\n", padByte);
	fprintf(stdout,"Output file: %s\n", fileName);
  
	if ((outFile = fopen(fileName, "wb")) == NULL) {
		fprintf(stderr,"Can't open file %s\n", fileName);
		exit(1);
	}
    
	if (lowAddr > highAddr) {
		lowAddr = 0;
		highAddr = 0;
	}
	fwrite(&binImage[lowAddr], 1, highAddr-lowAddr, outFile);
	free(binImage);
	fclose(inFile);
	fclose(outFile);
	return 0;
}
